#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "mesh.h"

#define PI 3.14159265

static GLUquadric *quadric;
static GLfloat mat_specular[] = { 1.0, 1.0, 1.0, 1.0 };
static GLfloat mat_shininess[] = { 90 };
static GLfloat light_position[] = { 1, 3, 2, 0};
static GLfloat light_color[] = { 1.0, 1.0, 1.0 };
static GLfloat light_direction[] = {0.0,0.0,-1.0};
static float time = 0;
static int wave = 0;
static Mesh mesh;
const int shellcount = 16;
static Mesh *shells[shellcount];
static GLuint textureID[2];


int keyPressed()
{
	SDL_Event event;
	SDL_PollEvent(&event);
	return event.type==SDL_KEYDOWN;
}

void distord(float *x, float *y, float *z)
{
	float t = 32*time + (*y)*8;
	if (!wave) return;
	(*x) += sin(t)*.2;
	(*z) += cos(t)*.2;
}

void calcSphereVertex(float x, float y)
{
	float px = sin(x)*cos(y), py = sin(y), pz = cos(x)*cos(y);
	distord(&px,&py,&pz);
	glVertex3f(px,py,pz);
}

void renderObject()
{
	float x,y,step=(2*PI)/64;

	glEnable(GL_NORMALIZE);
	glPushAttrib(GL_CURRENT_BIT);
	glPushMatrix();
	glScalef(.5,.5,.5);
	glTranslatef(0,1,0);
	glColor3f(.3,.1,.1);
//	glutSolidTeapot(0.5);

	for(y=0;y<2*PI-step;y+=step)
	{
		for(x=0;x<2*PI-step;x+=step)
		{
			glBegin(GL_QUADS);
			calcSphereVertex(x,y);
			calcSphereVertex(x+step,y);
			calcSphereVertex(x+step,y+step);
			calcSphereVertex(x,y+step);
			glNormal3f(sin(x)*cos(y),sin(y),cos(x)*cos(y));
			glEnd();
		}
	}

	glPopMatrix();
	glPopAttrib();
	glDisable(GL_NORMALIZE);
}

void mapCylindricalTexture(Vector *v, double scale)
{
	Vector p(*v*scale);
	glTexCoord2f( 4 * atan(p.getX()/p.getZ()) / (2*PI) , p.getY());
}

void renderMesh(Mesh *m)
{
	int i;

	glEnable(GL_NORMALIZE);
//	glColor3f(.4,.4,.4);

	glBegin(GL_TRIANGLES);
	for(i=0;i<m->faces;i++)
	{
		Mesh::Face *f = &m->face[i];
//		double txscale = 0.0004;
//		double txscale = 0.004;
		double txscale = 0.01;

		glNormal3f(f->normal[1].getX(),f->normal[1].getY(),f->normal[1].getZ());
//		glTexCoord2f(fabs(txscale*f->vertex[0]->position.getZ()),fabs(txscale*f->vertex[0]->position.getY()));
//		glTexCoord2f(f->texel[0][0],f->texel[0][1]);
		mapCylindricalTexture(&f->vertex[0]->position,txscale);
		glVertex3f(f->vertex[0]->position.getX(),f->vertex[0]->position.getY(),f->vertex[0]->position.getZ());

		glNormal3f(f->normal[2].getX(),f->normal[2].getY(),f->normal[2].getZ());
//		glTexCoord2f(fabs(txscale*f->vertex[1]->position.getZ()),fabs(txscale*f->vertex[1]->position.getY()));
//		glTexCoord2f(f->texel[1][0],f->texel[1][1]);
		mapCylindricalTexture(&f->vertex[1]->position,txscale);
		glVertex3f(f->vertex[1]->position.getX(),f->vertex[1]->position.getY(),f->vertex[1]->position.getZ());

		glNormal3f(f->normal[3].getX(),f->normal[3].getY(),f->normal[3].getZ());
//		glTexCoord2f(fabs(txscale*f->vertex[2]->position.getZ()),fabs(txscale*f->vertex[2]->position.getY()));
//		glTexCoord2f(f->texel[2][0],f->texel[2][1]);
		mapCylindricalTexture(&f->vertex[2]->position,txscale);
		glVertex3f(f->vertex[2]->position.getX(),f->vertex[2]->position.getY(),f->vertex[2]->position.getZ());


//		printf("%f, %f, %f\n",f->vertex[0]->position.getX(),f->vertex[0]->position.getY(),f->vertex[0]->position.getZ());
	}
	glEnd();
}

void renderFin(Mesh::Face *f, int a, int b)
{
	double extrusion = 10.0;
	Vector p, edge1, edge2;

	edge1 = f->vertex[b]->position - f->vertex[a]->position;
	edge2 = f->normal[a+1]*extrusion;
	
//	p = edge1.cross(edge2).normalize();
	glNormal3f(p.getX(),p.getY(),p.getZ());

	glColor4f(1.0,1.0,1.0,1.0);
	glTexCoord2f(0,1);
	glVertex3f(f->vertex[a]->position.getX(),f->vertex[a]->position.getY(),f->vertex[a]->position.getZ());
	glTexCoord2f(1,1);
	glVertex3f(f->vertex[b]->position.getX(),f->vertex[b]->position.getY(),f->vertex[b]->position.getZ());
	p = f->vertex[b]->position + f->normal[b+1]*extrusion;
	glTexCoord2f(1,0);
	glColor4f(1.0,1.0,1.0,0.0);
	glVertex3f(p.getX(),p.getY(),p.getZ());
	p = f->vertex[a]->position + f->normal[a+1]*extrusion;
	glTexCoord2f(0,0);
	glVertex3f(p.getX(),p.getY(),p.getZ());
}

void renderFins(Mesh *m)
{
	int i;
	glEnable(GL_NORMALIZE);
//	glColor3f(.4,.4,.4);
	glDepthMask(0);

	glBegin(GL_QUADS);
	for(i=0;i<m->faces;i++)
	{
		Mesh::Face *f = &m->face[i];

		renderFin(f,0,1);
		renderFin(f,1,2);
		renderFin(f,2,0);

/*
		glNormal3f(f->normal[1].getX(),f->normal[1].getY(),f->normal[1].getZ());
//		glTexCoord2f(fabs(txscale*f->vertex[0]->position.getZ()),fabs(txscale*f->vertex[0]->position.getY()));
//		glTexCoord2f(f->texel[0][0],f->texel[0][1]);
		mapCylindricalTexture(&f->vertex[0]->position,txscale);
		glVertex3f(f->vertex[0]->position.getX(),f->vertex[0]->position.getY(),f->vertex[0]->position.getZ());

		glNormal3f(f->normal[2].getX(),f->normal[2].getY(),f->normal[2].getZ());
//		glTexCoord2f(fabs(txscale*f->vertex[1]->position.getZ()),fabs(txscale*f->vertex[1]->position.getY()));
//		glTexCoord2f(f->texel[1][0],f->texel[1][1]);
		mapCylindricalTexture(&f->vertex[1]->position,txscale);
		glVertex3f(f->vertex[1]->position.getX(),f->vertex[1]->position.getY(),f->vertex[1]->position.getZ());

		glNormal3f(f->normal[3].getX(),f->normal[3].getY(),f->normal[3].getZ());
//		glTexCoord2f(fabs(txscale*f->vertex[2]->position.getZ()),fabs(txscale*f->vertex[2]->position.getY()));
//		glTexCoord2f(f->texel[2][0],f->texel[2][1]);
		mapCylindricalTexture(&f->vertex[2]->position,txscale);
		glVertex3f(f->vertex[2]->position.getX(),f->vertex[2]->position.getY(),f->vertex[2]->position.getZ());
*/

//		printf("%f, %f, %f\n",f->vertex[0]->position.getX(),f->vertex[0]->position.getY(),f->vertex[0]->position.getZ());
	}
	glDepthMask(1);
	glEnd();
}

void extrudeMesh(Mesh *m, double amount)
{
	int i;

	for(i=0;i<m->faces;i++)
	{
		Mesh::Face *f = &m->face[i];
		f->vertex[0]->position+=(f->normal[1]*(amount));
		f->vertex[1]->position+=(f->normal[2]*(amount));
		f->vertex[2]->position+=(f->normal[3]*(amount));
	}
}

void renderPlane()
{
	glPushAttrib(GL_CURRENT_BIT);
	glColor4f(.3,.3,.6,0.4);
	glBegin(GL_POLYGON);
	glVertex3f(-1.0,0,-1.0);
	glVertex3f( 1.0,0,-1.0);
	glVertex3f( 1.0,0, 1.0);
	glVertex3f(-1.0,0, 1.0);
	glEnd();
	glPopAttrib();
}

void draw()
{
	int i;
	double alpha, alphadelta = 0.6 / (double)(shellcount);
	GLfloat glass_specular[] = { 0.3, 0.3, 0.6, 0.3 };
	GLfloat glass_shininess[] = { 10.0 };

	time+=0.01;

	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT );

	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity( );

//	glRotatef(time, 0, 0, 1.0);
	glTranslatef( 0.0, 0.0, -3.0  + sin(time)*0);
	gluLookAt(0,0,0,sin(time/3)*4,sin(time/12.0)-2,cos(time/3)*4,0,1,0);


/*
	glEnable(GL_STENCIL_TEST);
	glStencilFunc(GL_ALWAYS,1,1);
	glStencilOp(GL_REPLACE,GL_REPLACE,GL_REPLACE);

	glDepthMask(0);
	glColorMask(0,0,0,0);
	renderPlane();
	glColorMask(1,1,1,1);
	glDepthMask(1);

	glPushMatrix();
	glStencilFunc(GL_EQUAL,1,1);
	glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
//	glRotatef(180,0,0,1);
//	glRotatef(180,0,1,0);
	glScalef(1,-1,1);
	glFrontFace(GL_CCW);
	wave=1;
	renderObject();
	wave=0;
	glFrontFace(GL_CW);
	glDisable(GL_STENCIL_TEST);
	glPopMatrix();

	renderPlane();
	renderObject();
*/
//	glTranslatef(0,1,0);
	glTranslatef(0,1,0);
//	glutSolidTeapot(0.5);

//	glScalef(0.0005,-0.0005,0.0005);	// duck
//	glScalef(0.005,0.005,0.005);		// torus
	glScalef(0.02,0.02,0.02);		// spring
	glColor4f(.7,.7,.7,1.0);

	glDisable(GL_TEXTURE_2D);
	renderMesh(&mesh);
	glEnable(GL_TEXTURE_2D);

#if 0
	glBindTexture(GL_TEXTURE_2D, textureID[0]);

	alpha=0.7;
	for(i=0;i<shellcount;i++,alpha-=alphadelta)
	{
		glPushMatrix();
		glColor4f(1.0,1.0,1.0,alpha);
		glTranslatef(cos(time*10)*(i/8.0),sin(time*10)*(i/8.0),0);
		renderMesh(shells[i]);
		glPopMatrix();
	}
#endif

#if 1
	glBindTexture(GL_TEXTURE_2D, textureID[1]);
	renderFins(&mesh);
#endif
}

void init(unsigned width, unsigned height)
{
	float ratio = (float)width / (float)height;
	int i;
	SDL_Surface *texture, *fin;

	glShadeModel(GL_SMOOTH);
	glCullFace( GL_BACK );
	glFrontFace( GL_CW );
	glEnable( GL_CULL_FACE );
	glClearColor( 0, 0, 0, 0 );
	glViewport( 0, 0, width, height );
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity( );
	gluPerspective( 60.0, ratio, 1.0, 1024.0 );
	quadric=gluNewQuadric();

        glLightfv(GL_LIGHT0, GL_POSITION, light_position);
        glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
        glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
        glLightfv(GL_LIGHT0, GL_AMBIENT, light_color);

	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
//	glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);
	glColorMaterial(GL_FRONT_AND_BACK,GL_DIFFUSE);

	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_COLOR_MATERIAL);
	glEnable(GL_TEXTURE);
	glEnable(GL_TEXTURE_2D);

        glShadeModel(GL_SMOOTH);

	texture = IMG_Load("hair.png");
	fin = IMG_Load("fin.png");

	memset(texture->pixels, 0, texture->w*texture->h*texture->format->BytesPerPixel);
	for(i=0;i<100000;i++)
	{
		((unsigned long*)(texture->pixels))[rand()%(texture->w*texture->h)] = -1;
	}

	printf("texture bits per pixel: %d\n",texture->format->BitsPerPixel);

	glGenTextures(sizeof(textureID)/sizeof(textureID[0]), (GLuint*)textureID);

	glBindTexture(GL_TEXTURE_2D, textureID[0]);
	gluBuild2DMipmaps(GL_TEXTURE_2D,4,texture->w,texture->h,GL_RGBA,GL_UNSIGNED_BYTE,texture->pixels);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	glBindTexture(GL_TEXTURE_2D, textureID[1]);
	gluBuild2DMipmaps(GL_TEXTURE_2D,4,fin->w,fin->h,GL_RGBA,GL_UNSIGNED_BYTE,fin->pixels);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
//	glEnable(GL_TEXTURE_GEN_S);
//	glEnable(GL_TEXTURE_GEN_T);
}

int main(int argc, char **argv)
{
	SDL_Surface *screen;
//	Uint16 *b = (Uint16*)screen;
	int i;
	double e;

	if (SDL_Init(SDL_INIT_AUDIO|SDL_INIT_VIDEO) < 0)
	{
	    	fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
    		exit(1);
	}
	atexit(SDL_Quit);

/*
	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
*/
	SDL_GL_SetAttribute( SDL_GL_STENCIL_SIZE, 8 );
//	screen = SDL_SetVideoMode(640,480,16,SDL_OPENGLBLIT);
	screen = SDL_SetVideoMode(640,480,32,SDL_OPENGL);

	SDL_GL_GetAttribute(SDL_GL_STENCIL_SIZE,&i);
	printf("stencil size %d.\n", i);

	mesh.load("duck.3ds");

	for(i=0,e=0.0;i<shellcount;i++,e+=0.1)
	{
		shells[i] = new Mesh(mesh);
		extrudeMesh(shells[i],e);
	}

	if (screen == NULL)
	{
		fprintf(stderr, "Unable to set 640x480 video: %s\n", SDL_GetError());
		exit(1);
	}

	init(640,480);

	while(!keyPressed())
	{
			draw();
			SDL_GL_SwapBuffers();
	}

	return 0;
}

