/***************************************************************************
                          mesh.cpp  -  description
                             -------------------
    begin                : Sun Nov 25 2001
    copyright            : (C) 2001 by Sami Kystil, Joonas Kerttula
    email                : hiteck@utanet.fi
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "mesh.h"
#include "ray.h"

Mesh::Mesh()
{
	face=NULL;
	vertex=NULL;
	lastFace=NULL;
	vertices=faces=0;
	root=NULL;
	id=0;
}

Mesh::Mesh(Mesh &b)
{
	int i;

	face=NULL;
	vertex=NULL;
	lastFace=NULL;
	vertices=faces=0;
	root=NULL;
	id=0;

	vertices = b.vertices;
	faces = b.faces;
	vertex = new Vertex[vertices];

	for(i=0;i<vertices;i++)
		vertex[i] = b.vertex[i];

	face = new Face[faces];
	for(i=0;i<faces;i++)
	{
		face[i] = b.face[i];
		face[i].vertex[0] = &vertex[(int)((b.face[i].vertex[0] - b.vertex))];
		face[i].vertex[1] = &vertex[(int)((b.face[i].vertex[1] - b.vertex))];
		face[i].vertex[2] = &vertex[(int)((b.face[i].vertex[2] - b.vertex))];
	}
//	buildTree(0,0);
}

bool Mesh::load(const char *source, const char *srcName, scalar scale, unsigned bspMinFaces, unsigned bspMaxDepth)
{
#ifdef HAVE_LIB3DS
	// this bit is almost straight from 3ds2m.c, which is a part of
	// ``The 3D Studio File Format Library'' by J.E. Hoffman <je-h@gmx.net>
	Lib3dsFile *f=NULL;
	Lib3dsMesh *m;
	unsigned i,j,k,l,n;
	Lib3dsVector pos;

	f=lib3ds_file_load(source);

	if (!f)
		return false;

	if (srcName)
	{
		for (m=f->meshes; m; m=m->next)
		{
			if (!strcmp(srcName, m->name))
				break;
		}
    if (!m)
      printf("No mesh named %s found in %s.\n",srcName,source);
	}
	else
		m=f->meshes;

	if (!m) return false;

//	Lib3dsVector *normal = new Lib3dsVector[3*m->faces];
//	lib3ds_mesh_calculate_normals(m, normal);

	vertices=m->points;
	vertex=new Vertex[m->points];
	if (!vertex) return false;

	faces=m->faces;
	face=new Face[m->faces];
	if (!face) return false;

	for(i=0; i<m->points; ++i)
	{
		lib3ds_vector_transform(pos, m->matrix, m->pointL[i].pos);
		vertex[i].position.set(pos[1]*scale,-pos[2]*scale,pos[0]*scale);
	}

	if (!m->texelL)
	{
		printf("Warning: 3DS file doesn't contain texture coordinates.\n");
	}

	for (i=0; i<m->faces; ++i)
	{
//		face[i].vertices=3;
//		face[i].vertex = new Vertex*[3];
		face[i].vertex[0] = &vertex[m->faceL[i].points[0]];
		face[i].vertex[1] = &vertex[m->faceL[i].points[1]];
		face[i].vertex[2] = &vertex[m->faceL[i].points[2]];

		if (m->texelL)
		{
			face[i].texel[0][0] = m->texelL[m->faceL[i].points[0]][0];
			face[i].texel[0][1] = m->texelL[m->faceL[i].points[0]][0];
			face[i].texel[1][0] = m->texelL[m->faceL[i].points[1]][0];
			face[i].texel[1][1] = m->texelL[m->faceL[i].points[1]][0];
			face[i].texel[2][0] = m->texelL[m->faceL[i].points[2]][0];
			face[i].texel[2][1] = m->texelL[m->faceL[i].points[2]][0];
		} else
		{
			face[i].texel[0][0] =
			face[i].texel[0][1] =
			face[i].texel[1][0] =
			face[i].texel[1][1] =
			face[i].texel[2][0] =
			face[i].texel[2][1] = 0;
		}

/*
//		face[i].normal[1] = Vector(normal[i*3+0][1],-normal[i*3+0][2],normal[i*3+0][0]);
//		face[i].normal[2] = Vector(normal[i*3+1][1],-normal[i*3+1][2],normal[i*3+1][0]);
//		face[i].normal[3] = Vector(normal[i*3+2][1],-normal[i*3+2][2],normal[i*3+2][0]);
		lib3ds_vector_transform(pos, m->matrix, normal[i*3+0]);
		face[i].normal[1] = Vector(pos[1],-pos[2],pos[0]).normalize();

//		printf("%f, %f, %f\n",pos[0],pos[1],pos[2]);
//		face[i].normal[1] = Vector(pos[0],pos[1],pos[2]);
		lib3ds_vector_transform(pos, m->matrix, normal[i*3+1]);
		face[i].normal[2] = Vector(pos[1],-pos[2],pos[0]).normalize();
//		face[i].normal[2] = Vector(pos[0],pos[1],pos[2]);
		lib3ds_vector_transform(pos, m->matrix, normal[i*3+2]);
		face[i].normal[3] = Vector(pos[1],-pos[2],pos[0]).normalize();
//		face[i].normal[3] = Vector(pos[0],pos[1],pos[2]);
*/
		prepareFace(&face[i]);
	}

//	delete normal;

#if 1	// group-based smoothing
	int **faceList = new (int*)[vertices];	// list of faces attached to a vertex
	unsigned int *faceCount = new unsigned int[vertices];	// number of faces attached to a vertex
	Vector smoothed[32];

	if (!faceList || !faceCount)
		return false;

	for(i=0; i<vertices; i++)
	{
		faceCount[i]=0;

		for(j=0; j<faces; j++)
			for(k=0; k<3; k++)
				if (face[j].vertex[k]==&vertex[i])
					faceCount[i]++;

//		printf("There are %d faces attached to vertex %d.\n",faceCount[i],i);
		faceList[i] = new int[faceCount[i]];
		faceCount[i]=0;

		for(j=0; j<faces; j++)
			for(k=0; k<3; k++)
				if (face[j].vertex[k]==&vertex[i])
					faceList[i][faceCount[i]++] = j;
/*
		Vector normal = Vector(0,0,0);
		for(j=0; j<faces; j++)
			for(k=0; k<face[j].vertices; k++)
			{
				if (face[j].vertex[k]==&vertex[i])
				{
					normal+=face[j].normal[0];
					n++;
				}
			}
		normal.normalize();
		for(j=0; j<faces; j++)
			for(k=0; k<face[j].vertices; k++)
			{
				if (face[j].vertex[k]==&vertex[i])
				{
					face[j].normal[k+1] = normal;
					n++;
				}
			}
*/
	}
#endif

#if 1	// smooth all faces
	for(i=0;i<vertices;i++)
	{
		vertex[i].normal = Vector(0,0,0);
		for(j=0; j<faceCount[i]; j++)
			vertex[i].normal+=face[faceList[i][j]].normal[0];
		vertex[i].normal.normalize();
	}

	for(j=0; j<faces; j++)
		for(k=0; k<3; k++)
			face[j].normal[k+1] = face[j].vertex[k]->normal;
/*
	for(i=0;i<vertices;i++)
	{
		Vector normal = Vector(0,0,0);
		for(j=0; j<faceCount[i]; j++)
			normal+=face[faceList[i][j]].normal[0];
		normal.normalize();

		for(j=0; j<faces; j++)
			for(k=0; k<face[j].vertices; k++)
			{
				if (face[j].vertex[k]==&vertex[i])
				{
					face[j].normal[k+1] = normal;
					n++;
				}
			}
	}
*/
#else	// group-based smoothing part two
	for(i=0; i<faces; i++)
	{
		Lib3dsFace *protoFace = &m->faceL[i];
		
		if (protoFace->smoothing)
		{
			for(j=0; j<3; j++)
			{
//				int v=-1;
				int v=protoFace->points[j];
				n=0;
/*				
				for(k=0;k<vertices;k++)
					if (face[i].vertex[j] == &vertex[k])
					{
						v=k;
						break;
					}
				if (v<0) continue;
*/				
				face[i].normal[j+1]=Vector(0,0,0);
				
				for(k=0; k<faceCount[v]; k++)
				{
					int faceN = faceList[v][k];
					Lib3dsFace *f = &m->faceL[faceN];

//					printf("%08X <-> %08X\n",protoFace->smoothing, f->smoothing);
					if (protoFace->smoothing & f->smoothing)
					{
						bool found=false;
						for(l=0;l<n;l++)
							if (smoothed[l].dot(face[faceN].normal[0]) < 1e-5)
							{
								found=true;
								break;
							}
						if (!found)
						{
//							printf("smoothing face %d / vertex %d with face %d\n",i,j,faceN);
							face[i].normal[j+1]+=face[faceN].normal[0];
							smoothed[n++]=face[faceN].normal[0];
							if (n>=sizeof(smoothed)/sizeof(smoothed[0])) break;
						}
					}
				}
				face[i].normal[j+1].normalize();
			}
		}
		else
			face[i].normal[1]=face[i].normal[2]=face[i].normal[3]=face[i].normal[0];
	}
#endif		

	for(i=0;i<vertices;i++)
		delete[] faceList[i];
	delete[] faceList;
	delete[] faceCount;
	
	lib3ds_file_free(f);
	
/*	
	for (i=0;i<m->faces;++i)
	{
		Lib3dsFace *f=&m->faceL[i];
		for (j=0;j<3;++j)
		{
			Lib3dsVector n,N[32];
			Lib3dsFaces *p;
			int k,l;
			int found;

			if (f->smoothing)
			{
				lib3ds_vector_zero(n);
				k=0;
				for (p=fl[f->points[j]]; p; p=p->next)
				{
					found=0;
					for (l=0; l<k; ++l)
					{
						if (fabs(lib3ds_vector_dot(N[l], p->face->normal)-1.0)<1e-5)
						{
							found=1;
							break;
						}
					}
					if (!found)
					{
						if (f->smoothing & p->face->smoothing)
						{
							lib3ds_vector_add(n,n, p->face->normal);
							lib3ds_vector_copy(N[k], p->face->normal);
							++k;
						}
					}
				}
			}
			else
			{
				lib3ds_vector_copy(n, f->normal);
			}
			lib3ds_vector_normalize(n);
			lib3ds_vector_copy(normalL[3*i+j], n);
		}
	}
*/

#endif
	prepareBoundingObject();

	buildTree(bspMinFaces,bspMaxDepth);

	printf("Mesh %s loaded from %s (%d faces, %d vertices).\n",srcName,source,faces,vertices);
	
	return true;
}

void Mesh::prepareBoundingObject()
{
	unsigned i;
	bounding_radius_squared=0.0;

	scalar xmin,xmax,ymin,ymax,zmin,zmax;
	xmin = ymin = zmin = 1e30;
	xmax = ymax = zmax = -1e30;

	for(i=0; i<vertices; i++)
	{
		scalar x = vertex[i].position.getX();
		scalar y = vertex[i].position.getY();
		scalar z = vertex[i].position.getZ();
		if (x > xmax)
			xmax = x;
		if (x < xmin)
			xmin = x;
		if (y > ymax)
			ymax = y;
		if (y < ymin)
			ymin = y;
		if (z > zmax)
			zmax = z;
		if (z < zmin)
			zmin = z;
	}
	bounding_origin = Vector((xmax+xmin)*.5,(ymax+ymin)*.5,(zmax+zmin)*.5);
	for(i=0; i<vertices; i++)
	{
    Vector v = vertex[i].position - bounding_origin;
		scalar d = v.dot(v);
		if (d > bounding_radius_squared)
			bounding_radius_squared = d;
	}
}

Mesh::~Mesh()
{
/*	
	unsigned i;
	for(i=0;i<faces;i++)
	{
		delete[] face[i].vertex;
		delete[] face[i].plane;
	}
*/
	if (face) delete[] face;
	if (face) delete[] vertex;
	
	face=NULL;
	vertex=NULL;
	lastFace=NULL;

	if (root)
		delete root;
	root=NULL;
	
	vertices=faces=0;
}

void Mesh::prepareFace(Face *face)
{
	Vector *v1, *v2, *v3;
	
	v1=&face->vertex[0]->position;
	v2=&face->vertex[1]->position;
	v3=&face->vertex[2]->position;
	face->normal[0] = -(*v2-*v1).cross(*v3-*v2).normalize();
	face->edge[0] = *v2 - *v1;
	face->edge[1] = *v3 - *v1;
	face->id = 0;
	face->t = 0;

//	printf("normal: %f, %f, %f\n",face->normal.getX(),face->normal.getY(),face->normal.getZ());

/*	
	unsigned i;
	face->plane = new Plane[face->vertices+1];
	face->plane[0].normal = face->normal[0];
	face->plane[0].distance = -face->normal[0].dot(face->vertex[0]->position);
	
	v1=&face->vertex[face->vertices-1]->position;
	for(i=0;i<face->vertices;i++)
	{
		v2=&face->vertex[i]->position;
		face->plane[i+1].normal = ((*v2-*v1).cross(face->normal[0])).normalize();
		face->plane[i+1].distance = -face->plane[i+1].normal.dot(*v1);
		v1=v2;
	}
*/
}

scalar Mesh::intersect(Ray &ray)
{
	// test intersection with the bounding sphere first.
  Vector o = ray.origin - bounding_origin;
//  Vector o = ray.origin;
//  Vector o = ray.origin + trans.getOrigin();
	scalar a = ray.direction.dot(ray.direction);
	scalar b = 2*o.dot(ray.direction);
	scalar c = o.dot(o) - bounding_radius_squared;
//	scalar c = o.dot(o) - 1000;
	scalar d = b*b - 4*a*c;

#if 0 // render bounding spheres
  scalar t;
  lastFace=&face[0];

	if (d>1e-4)
	{
		d = sqrt(d);
		t = 0.5 * ((-b > d + 1e-4? -d : d) - b)/a;
		return t;
  }
	else
		return -1;
#endif

	if (d<0)
		return -1;

//	intersections=0;

#if 1 // Find intersection with binary space partition tree (BSP)
	id++;
	lastFace=NULL;
	// reference
	return root->intersect(this, ray.origin, ray.direction, 0, 1e30);

//	printf("%03.2f%%\t - %d/%d intersections\n",100.0*(scalar)intersections/(scalar)faces,intersections,faces);

//	return t;
#else
	unsigned i, j;
	scalar nearest=1e30;

	lastFace=NULL;
	for(i=0;i<faces;i++)
	{
//		This algorithm adapted from the paper
//    ``Fast, minimum storage ray-triangle intersection''
//    by Tomas Mller and Ben Trumbore.
//    (Journal of Graphics Tools, 2(1):21--28, 1997)
//
//    http://www.graphics.cornell.edu/pubs/1997/MT97.html
		Face *f = &face[i];
		scalar t = intersectFace(ray.origin,ray.direction,f);

		// see if its the nearest hit found so far, and if so, remember it
		if (t>1e-4 && t<nearest)
		{
			nearest=t;
			lastFace=f;
		}

#if 0
		// this plane-intersection polygon test code is obsolete
		scalar div = (face[i].plane[0].normal.dot(ray.direction));
		if (fabs(div)>1e-4)
		{
			t=-(face[i].plane[0].distance+face[i].plane[0].normal.dot(o))/div;
			if (t>1e-3)
			{
				bool inside=true;
				Vector intersection = o+ray.direction*t;
/*
				scalar s,t,r;
				Vector a,b,c,n;

				a = face[i].vertex[1]->position - face[i].vertex[0]->position;
				b = face[i].vertex[2]->position - face[i].vertex[0]->position;
				n = face[i].normal[0];
//				c = face[i].vertex[0]->position -intersection;
//				s = c.cross(b).dot(face[i].normal[0]);
//				t = a.cross(c).dot(face[i].normal[0]);

//				s = ((c.cross(b)).dot(n)) / n.length();
//				t = ((a.cross(c)).dot(n)) / n.length();
//				r = 1 - (s + t);

				printf("s:%5.5f ",s);
				printf("t:%5.5f ",t);
				printf("r:%5.5f ",r);
				printf("\n");

				if (s<0 || t<0 || s>1 || t>1)
					continue;
*/			
				for(j=1;j<face[i].vertices+1;j++)
				{
					if ((face[i].plane[j].normal.dot(intersection)+face[i].plane[j].distance) < 0)
					{
						inside=false;
						break;
					}
				}
				if (!inside) continue;
					
				if (t<nearest2)
				{
					nearest2=t;
					lastFace=&face[i];
				}
			}
		}
#endif
	}

	printf("%d intersections\n",intersections);

	if (lastFace)
		return nearest;
	return -1;
#endif
}

Vector Mesh::normal(Vector &point)
{
	// XXX: assumes triangles
//	Vector		p = point - trans.getOrigin();	
	Vector normal =
		lastFace->normal[2] * lastFace->u +
		lastFace->normal[3] * lastFace->v +
		lastFace->normal[1] * lastFace->w;
/*

	Vector		*a = &point, *b, *c;
	Vector		normal;
	unsigned	i;
	scalar		area = (lastFace->vertex[1]->position-lastFace->vertex[0]->position).cross(lastFace->vertex[2]->position-lastFace->vertex[0]->position).length();

	b=&lastFace->vertex[2]->position;
	for(i=0;i<3;i++)
	{
		c=&lastFace->vertex[i]->position;
		scalar factor = ((*b-*a).cross(*c-*a).length()) / area;
//		normal+=lastFace->vertex[(i+1)%3]->normal * factor;
		normal+=lastFace->normal[(i+1)%3+1] * factor;
		b=c;	
	}

	normal-=normal2;
	printf("%f, %f, %f",normal.getX(),normal.getY(),normal.getZ());
*/
	return normal.normalize();
}

Mesh::Node::Node(unsigned _faces, Face **_face, unsigned _minFaces, unsigned _maxDepth)
{
	faces=_faces;
	minFaces=_minFaces;
	maxDepth=_maxDepth;
	if (faces)
	{
		face = new Face*[faces];
		memcpy(face,_face,sizeof(Face*)*faces);
	}
	front=back=NULL;
}

Mesh::Node::~Node()
{
	if (face)
		delete[] face;
	if (front)
		delete front;
	if (back)
		delete back;

	faces=0;
	face=NULL;
	front=back=NULL;
}

void Mesh::Node::split(unsigned depth, unsigned _axis)
{
	unsigned i,cfront,cback;
	Face **lfront, **lback;
	axis=_axis;

	if (faces<minFaces || depth>maxDepth)
		return;
	
	position=0;
	for(i=0;i<faces;i++)
	{
		position+=face[i]->vertex[0]->position.get()[axis];
		position+=face[i]->vertex[1]->position.get()[axis];
		position+=face[i]->vertex[2]->position.get()[axis];
	}
	position/=(faces*3);

	lfront=new Face*[faces];
	lback=new Face*[faces];

	cfront=cback=0;
	for(i=0;i<faces;i++)
		switch(classify(face[i]))
		{
		case 0:
			lfront[cfront++]=face[i];
		break;
		case 1:
			lback[cback++]=face[i];
		break;
		default:
			lfront[cfront++]=lback[cback++]=face[i];
		}

	if (cfront!=faces && cback!=faces)
	{
		printf("depth %d: %d initial faces, after split %d front, %d back.\n",depth,faces,cfront,cback);

		front=new Node(cfront,lfront,minFaces,maxDepth);
		delete[] lfront;	
		back=new Node(cback,lback,minFaces,maxDepth);
		delete[] lback;

		_axis = (_axis+1)%3;
		front->split(depth+1,_axis);
		back->split(depth+1,_axis);

		faces=0;
		delete[] face;
		face=NULL;
	}
}

unsigned Mesh::Node::classify(Vector *v)
{
	return (v->get()[axis] >= position) ? 0:1;
}

unsigned Mesh::Node::classify(Face *f)
{
	unsigned a = classify(&f->vertex[0]->position);
	unsigned b = classify(&f->vertex[1]->position);
	unsigned c = classify(&f->vertex[2]->position);

	if (!a && !b && !c)
		return 0;
	if (a && b && c)
		return 1;
	return 2;
}

scalar Mesh::intersectFace(Vector origin, Vector direction, Face *f)
{
//	intersections++;
//		This algorithm adapted from the paper
//    ``Fast, minimum storage ray-triangle intersection''
//    by Tomas Mller and Ben Trumbore.
//    (Journal of Graphics Tools, 2(1):21--28, 1997)
//
//    http://www.graphics.cornell.edu/pubs/1997/MT97.html
	Vector tvec, pvec, qvec;
	scalar det, inv_det;
	scalar t;

	// begin calculating determinant -- also used to calculate U parameter.
	pvec = direction.cross(f->edge[1]);
	det = f->edge[0].dot(pvec);

	// if determinant is near zero, the ray lies on the triangle plane
	if (det > -1e-6 && det < 1e-6)
		return -1;
	inv_det = 1.0/det;

	// calculate distance from the first vertex to the ray origin
	tvec = origin - f->vertex[0]->position;

	// calculate the U parameter and test bounds
	f->u = tvec.dot(pvec) * inv_det;
	if (f->u < 0.0 || f->u > 1.0)
		return -1;

	// prepare to test the V parameter
	qvec = tvec.cross(f->edge[0]);

	// calculate the V parameter and test bounds
	f->v = direction.dot(qvec) * inv_det;
	f->w = 1.0 - (f->u+f->v);
	if (f->v < 0.0 || f->w < 0.0)
		return -1;

	// calculate the ray parameter
	t = f->edge[1].dot(qvec) * inv_det;

	/*
	       /\
	      /| \
	     / |  \
	    /v |  w\
	   / _-+-__ \
	  /-   v   --\
	 --------------

	*/
	
	// u,v and w are now the areas of the three subtriangles (formed by
	// the intersection point and the triangle edges) divided by the whole
	// triangle area. This information is used in Mesh::normal() to calculate
	// the interpolated normal.

	return (t>1e-4)?t:-1;
}

scalar Mesh::Node::intersect(Mesh *m, Vector origin, Vector direction, scalar t_min, scalar t_max)
{
	scalar div, t=-1;
	Node *near, *far;

	if (faces)
	{
		unsigned i;
		scalar nearest=1e30;

		m->lastFace=NULL;

		for(i=0;i<faces;i++)
		{
			Face *f = face[i];

			if (f->id!=m->id)
			{
				f->t = m->intersectFace(origin,direction,f);
				f->id=m->id;
			}

			if (f->t!=-1)
			{
				if (f->t>t_min && f->t<t_max && f->t<nearest)
				{
					nearest=f->t;
					m->lastFace=f;
				}
			}
		}
		return m->lastFace?nearest:-1;
	}

	div = direction.get()[axis];

	if (fabs(div)>1e-3)
//		return -1;
//		t = -(origin.get()[axis] + position)/div;
		t = (position - origin.get()[axis])/div;

//	Vector a = origin + direction*t_min;
//	Vector b = origin + direction*t_max;
//	unsigned sa = classify(&a);
//	unsigned sb = classify(&b);

	if (classify(&origin)==0)
	{
		near=front;
		far=back;
	}
	else
	{
		near=back;
		far=front;
//		t=-t;
	}

//	printf("%d intersect: %f\n",m->id,t);
//	printf("%f, %f, %d, %d\n",t_min,t_max,sa,sb);

//	if (t_min>t && t_max>t)
//	if (t<1e-3 || (sa==0 && sb==0))
	if (t>t_max || t<0)
		return near->intersect(m, origin, direction, t_min, t_max);
//	else if (t_min<=t && t_max<=t)
//	else if (sa==1 && sb==1)
	else if (t<t_min)
		return far->intersect(m, origin, direction, t_min, t_max);
	else
	{
		scalar t0 = near->intersect(m, origin, direction, t_min, t);
		if (t0!=-1) return t0;
		return far->intersect(m, origin, direction, t, t_max);
	}
}

void Mesh::buildTree(unsigned minFaces, unsigned maxDepth)
{
	unsigned i;
	Face **f = new Face*[faces];

	if (!minFaces) minFaces=8;
	if (!maxDepth) maxDepth=32;

	for(i=0;i<faces;i++)
		f[i]=&face[i];

	root = new Node(faces,f,minFaces,maxDepth);
	delete[] f;
	root->split(0,0);
}

void Mesh::derivatives(Vector &u, Vector &v)
{
//	u = lastFace->u;
//	v = lastFace->v;
}
