/***************************************************************************
                          mesh.h  -  description
                             -------------------
    begin                : Sun Nov 25 2001
    copyright            : (C) 2001 by Sami Kystil, Joonas Kerttula
    email                : hiteck@utanet.fi
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef MESH_H
#define MESH_H

#define HAVE_LIB3DS

#ifdef HAVE_LIB3DS
#include <lib3ds/file.h>
#include <lib3ds/mesh.h>
#include <lib3ds/vector.h>
#endif

#include "vector.h"

class Ray;

/**
  *@author Sami Kystil, Joonas Kerttula
  */

class Mesh
{
public:
	Mesh();
	Mesh(Mesh &b);
	~Mesh();

	scalar	intersect(Ray &ray);
	Vector	normal(Vector &point);
	bool		load(const char *source, const char *srcName=NULL, scalar scale=1.0, unsigned bspMaxFaces=0, unsigned bspMaxDepth=0);
	void		derivatives(Vector &u, Vector &v);

	typedef struct
	{
		scalar		distance;
		Vector		normal;
	} Plane;

	typedef struct
	{
		Vector		position;
		Vector		normal;
	} Vertex;

	// XXX: currently limited to triangles
	typedef struct
	{
//		unsigned	vertices;
//		Vertex		**vertex;
		Vertex		*vertex[3];
//		Plane			*plane;			// first plane is the plane of the face itself, the rest are the edge planes.
		Vector		normal[4];	// the first one is the normal of the face, the rest are vertex normals.
		Vector		edge[2];
		scalar		texel[3][2];
		unsigned	id;
		scalar		t;
		scalar		u,v,w;
	} Face;

	// a single node in the binary space partition tree
	class Node
	{
	public:
		Node(unsigned _faces, Face **_face, unsigned _minFaces, unsigned _maxDepth);
		~Node();
		// recursively splits the set of polygons into two roughly equal partitions and create a subnode for each half.
		void				split(unsigned depth, unsigned _axis);
		// given a vector, determines which side of this partition it originates from
		unsigned		classify(Vector *v); // 0: front, 1: back
		// performs the classification on a face
		unsigned		classify(Face *f);	// 0: front, 1: back, 2:both
		// recursively intersects a ray with the polygons contained within this node, or,
		// if this is a leaf node, both subnodes.
		scalar			intersect(Mesh *m, Vector origin, Vector direction, scalar t_min, scalar t_max);

		unsigned		faces;
		Face				**face;
		unsigned		axis, minFaces, maxDepth;
		scalar			position;
		class Node	*front, *back;
	};

	// performs vector-polygon intersection calculation
	scalar	intersectFace(Vector origin, Vector direction, Face *f);
	// initializes a face structure
	void		prepareFace(Face *face);
	// prepares an optimal bounding object for the current mesh
	void		prepareBoundingObject();
	// initializes the binary space partition tree
	void		buildTree(unsigned minFaces, unsigned maxDepth);

	unsigned		faces, vertices, id, intersections;
	Vertex			*vertex;
	Face				*face, *lastFace;
	scalar			bounding_radius_squared;
	Vector			bounding_origin;
	Node				*root;
friend class Mesh::Node;
};

#endif
