#ifndef VECTOR_H
#define VECTOR_H

// Vector math stuff (C) 2001 by Sami Kystil

#include <math.h>

//const PI = 3.1415926535897932384626433832795028841971693993751f;
const double PI = 3.141592653;
typedef double scalar;

class Vector	// 3
{
friend class Matrix;
public:
//	Vector() {v[0]=v[1]=v[2]=0;}
	Vector() {}
	Vector(const scalar _x, const scalar _y, const scalar _z) {v[0]=_x;v[1]=_y;v[2]=_z;}
	Vector(const class Vector &b) {v[0]=b.v[0]; v[1]=b.v[1]; v[2]=b.v[2]; }
	Vector operator=(const class Vector &b)  {v[0]=b.v[0];v[1]=b.v[1];v[2]=b.v[2]; return *this;}
	inline class Vector operator+(const class Vector &b) const
	{
		return Vector(v[0]+b.v[0],v[1]+b.v[1],v[2]+b.v[2]);
	}
	inline scalar operator[](int i) const
	{
		return v[i];
	}
	inline Vector operator+=(const class Vector &b)
	{
		v[0]+=b.v[0]; v[1]+=b.v[1]; v[2]+=b.v[2];
		return *this;
	}
	inline Vector operator-=(const class Vector &b)
	{
		v[0]-=b.v[0]; v[1]-=b.v[1]; v[2]-=b.v[2];
		return *this;
	}
	inline Vector operator*=(const scalar s)
	{
		v[0]*=s; v[1]*=s; v[2]*=s;
		return *this;
	}
	inline Vector operator/=(const scalar s)
	{
		scalar f=1.0/s;
		v[0]*=f;
		v[1]*=f;
		v[2]*=f;
		return *this;
	}
	inline Vector operator-(const class Vector &b) const
	{
		return Vector(v[0]-b.v[0],v[1]-b.v[1],v[2]-b.v[2]);
	}
	inline Vector operator-() const
	{
		return Vector(-v[0],-v[1],-v[2]);
	}
	inline Vector operator*(class Vector &b) const
	{
		return Vector(v[0]*b.v[0],
									v[1]*b.v[1],
									v[2]*b.v[2]);
	}
	inline Vector operator/(const scalar f) const
	{
		scalar p=1/f;
		return Vector(v[0]*p,
									v[1]*p,
									v[2]*p);
	}
	inline Vector operator*(const scalar f) const
	{
		return Vector(v[0]*f,v[1]*f,v[2]*f);
	}
	inline scalar dot(const Vector &b) const
	{
		return (scalar)(
			v[0]*b.v[0]+
			v[1]*b.v[1]+
			v[2]*b.v[2]);
	}
	inline Vector cross(const class Vector &b) const
	{
		return Vector(
			v[1]*b.v[2] - b.v[1]*v[2],
			v[2]*b.v[0] - b.v[2]*v[0],
			v[0]*b.v[1] - b.v[0]*v[1]);
	}
	scalar length() const
	{
		return (scalar)sqrt(v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
	}
	scalar lengthSquared() const
	{
		return (v[0]*v[0]+v[1]*v[1]+v[2]*v[2]);
	}
	Vector normalize()
	{
		scalar l=length();

		if (l)
			(*this)*=(1.0/l);
		
		return *this;
	}
	inline void set(const scalar _x, const scalar _y, const scalar _z) {v[0]=_x;v[1]=_y;v[2]=_z;}

	inline scalar getX() const {return v[0];}
	inline scalar getY() const {return v[1];}
	inline scalar getZ() const {return v[2];}
	scalar *get() {return v;}
	inline void setX(scalar _x) {v[0]=_x;}
	inline void setY(scalar _y) {v[1]=_y;}
	inline void setZ(scalar _z) {v[2]=_z;}
	
	inline bool nonzero() const { return (v[0]||v[1]||v[2]); }
	inline bool positive() const { return (v[0]>=0&&v[1]>=0&&v[2]>=0); }
protected:
	scalar	v[3];
};

class Quad
{
public:
	Quad() {v[0]=v[1]=v[2]=v[3]=0;}
	Quad(scalar _x, scalar _y, scalar _z, scalar _w) {v[0]=_x;v[1]=_y;v[2]=_z;v[3]=_w;}
	scalar *get() {return v;}
private:
	scalar	v[4];
};

#define M_CELL(col,row) (((row)<<2)+(col))
class Matrix	// 4x4
{
public:
	Matrix(Matrix *b)
	{
		memcpy(&m[0],&b->m[0],sizeof(m));
	}
	Matrix() {makeIdentity();}
//	scalar *get() {return &m[0];}
	scalar &cell(int row, int column) {return m[M_CELL(column,row)];}
	Matrix makeIdentity()
	{
		m[M_CELL(0,0)]=1.0;		m[M_CELL(1,0)]=0.0;		m[M_CELL(2,0)]=0.0;		m[M_CELL(3,0)]=0.0;
		m[M_CELL(0,1)]=0.0;		m[M_CELL(1,1)]=1.0;		m[M_CELL(2,1)]=0.0;		m[M_CELL(3,1)]=0.0;
		m[M_CELL(0,2)]=0.0;		m[M_CELL(1,2)]=0.0;		m[M_CELL(2,2)]=1.0;		m[M_CELL(3,2)]=0.0;
		m[M_CELL(0,3)]=0.0;		m[M_CELL(1,3)]=0.0;		m[M_CELL(2,3)]=0.0;		m[M_CELL(3,3)]=1.0;
		return *this;
	}
	inline Matrix operator*=(Matrix b)
	{
		(*this) = (*this) * b;
		return *this;
	}
	inline Vector getTranslation() const
	{
		return Vector(
			m[M_CELL(3,0)],
			m[M_CELL(3,1)],
			m[M_CELL(3,2)]);
	}
	inline Matrix transpose3x3() const
	{
		Matrix n;
		unsigned i,j;

		for(i=0;i<3;i++)
			for(j=0;j<3;j++)
				n.m[M_CELL(j,i)] = m[M_CELL(i,j)];
/*
		for(i=0;i<3;i++)
			n.m[M_CELL(3,i)] =
				m[M_CELL(i,0)] * m[M_CELL(3,0)] -
				m[M_CELL(i,1)] * m[M_CELL(3,1)] -
				m[M_CELL(i,2)] * m[M_CELL(3,2)];
*/
		return n;
	}
	inline Matrix operator*(Matrix b) const
	{
		int i,j;
		Matrix n;

		for (i=0;i<4;i++)
		for (j=0;j<4;j++)
		{
			n.m[M_CELL(i,j)] =
				m[M_CELL(i,0)]*b.m[M_CELL(0,j)] +
				m[M_CELL(i,1)]*b.m[M_CELL(1,j)] +
				m[M_CELL(i,2)]*b.m[M_CELL(2,j)] +
				m[M_CELL(i,3)]*b.m[M_CELL(3,j)];
		}
		return n;
	}
	inline Vector mul3x3(Vector v) const
	{
		return Vector(
			v.v[0]*m[M_CELL(0,0)] + v.v[1]*m[M_CELL(1,0)] + v.v[2]*m[M_CELL(2,0)],
			v.v[0]*m[M_CELL(0,1)] + v.v[1]*m[M_CELL(1,1)] + v.v[2]*m[M_CELL(2,1)],
			v.v[0]*m[M_CELL(0,2)] + v.v[1]*m[M_CELL(1,2)] + v.v[2]*m[M_CELL(2,2)]);
	}
	inline Vector operator*(Vector v) const
	{
		return Vector(
			v.v[0]*m[M_CELL(0,0)] + v.v[1]*m[M_CELL(1,0)] + v.v[2]*m[M_CELL(2,0)] + m[M_CELL(3,0)],
			v.v[0]*m[M_CELL(0,1)] + v.v[1]*m[M_CELL(1,1)] + v.v[2]*m[M_CELL(2,1)] + m[M_CELL(3,1)],
			v.v[0]*m[M_CELL(0,2)] + v.v[1]*m[M_CELL(1,2)] + v.v[2]*m[M_CELL(2,2)] + m[M_CELL(3,2)]);
	}
	inline Matrix operator*(scalar s) const
	{
		int i,j;
		Matrix n;

		for (i=0;i<4;i++)
		for (j=0;j<4;j++)
			n.m[M_CELL(i,j)] = m[M_CELL(i,j)]*s;
		return n;
	}
	inline void setRow(int row, Vector v)
	{
		m[M_CELL(0,row)]=v.getX();
		m[M_CELL(1,row)]=v.getY();
		m[M_CELL(2,row)]=v.getZ();
	}
	Vector getRow(int row)
	{
		return class Vector(m[(row<<2)+0],m[(row<<2)+1],m[(row<<2)+2]);
	}
	inline Matrix makeTranslation(Vector direction) const
	{
		Matrix	n;
		n.m[M_CELL(3,0)] = direction.v[0];
		n.m[M_CELL(3,1)] = direction.v[1];
		n.m[M_CELL(3,2)] = direction.v[2];
		return n;
	}
	inline Matrix makeRotation(Vector axis, scalar angle) const
	{
		scalar	s, c, t;
		Matrix	n;

		axis = axis.normalize();
		s = sin(angle);
		c = cos(angle);
		t = 1.0 - c;

		n.m[M_CELL(0,0)] = t * axis.v[0] * axis.v[0] + c;
		n.m[M_CELL(1,0)] = t * axis.v[0] * axis.v[1] + s * axis.v[2];
		n.m[M_CELL(2,0)] = t * axis.v[0] * axis.v[2] - s * axis.v[1];
		n.m[M_CELL(3,0)] = 0;
		n.m[M_CELL(0,1)] = t * axis.v[0] * axis.v[1] - s * axis.v[2];
		n.m[M_CELL(1,1)] = t * axis.v[1] * axis.v[1] + c;
		n.m[M_CELL(2,1)] = t * axis.v[1] * axis.v[2] + s * axis.v[0];
		n.m[M_CELL(3,1)] = 0;
		n.m[M_CELL(0,2)] = t * axis.v[0] * axis.v[2] + s * axis.v[1];
		n.m[M_CELL(1,2)] = t * axis.v[1] * axis.v[2] - s * axis.v[0];
		n.m[M_CELL(2,2)] = t * axis.v[2] * axis.v[2] + c;
		n.m[M_CELL(3,2)] = 0;
		n.m[M_CELL(0,3)] = 0;
		n.m[M_CELL(1,3)] = 0;
		n.m[M_CELL(2,3)] = 0;
		n.m[M_CELL(3,3)] = 1;

		return n;
	}
	inline Matrix inverse3x3() const
	{
			Matrix n;
			scalar det;
			det = m[0] * ( m[4]*m[8] - m[7]*m[5] )
          - m[1] * ( m[3]*m[8] - m[6]*m[5] )
          + m[2] * ( m[3]*m[7] - m[6]*m[4] );

     if ( fabs( det ) < 0.0005 )
       {
				return n;
       }

     n.m[0] =    m[4]*m[8] - m[5]*m[7]   / det;
     n.m[1] = -( m[1]*m[8] - m[7]*m[2] ) / det;
     n.m[2] =    m[1]*m[5] - m[4]*m[2]   / det;

     n.m[3] = -( m[3]*m[8] - m[5]*m[6] ) / det;
     n.m[4] =    m[0]*m[8] - m[6]*m[2]   / det;
     n.m[5] = -( m[0]*m[5] - m[3]*m[2] ) / det;

     n.m[6] =    m[3]*m[7] - m[6]*m[4]   / det;
     n.m[7] = -( m[0]*m[7] - m[6]*m[1] ) / det;
     n.m[8] =    m[0]*m[4] - m[1]*m[3]   / det;

		return n;

/*
		// thanks to Miika Sell [itchi] for this!
		Matrix n;
//		scalar mat[]={m[data[0][0],data[0][1],data[0][2],data[1][0],data[1][1],data[1][2],data[2][0],data[2][1],data[2][2]};
//		scalar mat[]={m[M_CELL(0,0)],m[M_CELL(0,1)],m[M_CELL(0,2)],m[M_CELL(1,0)],m[M_CELL(1,1)],m[M_CELL(1,2)],m[M_CELL(2,0)],m[M_CELL(2,1)],m[M_CELL(2,2)]};
 		scalar mat[]={m[M_CELL(0,0)],m[M_CELL(1,0)],m[M_CELL(2,0)],m[M_CELL(0,1)],m[M_CELL(1,1)],m[M_CELL(2,1)],m[M_CELL(0,2)],m[M_CELL(1,2)],m[M_CELL(2,2)]};
		scalar D=(mat[0]*mat[4]*mat[8] + mat[1]*mat[5]*mat[6] + mat[2]*mat[3]*mat[7]) -
							(mat[0]*mat[5]*mat[7] + mat[1]*mat[3]*mat[8] + mat[2]*mat[4]*mat[6]);
		scalar inv[9];

		inv[0]=mat[4]*mat[8]-mat[5]*mat[7];
		inv[1]=mat[2]*mat[7]-mat[1]*mat[8];
		inv[2]=mat[1]*mat[5]-mat[2]*mat[4];
		inv[3]=mat[5]*mat[6]-mat[3]*mat[8];
		inv[4]=mat[0]*mat[8]-mat[2]*mat[6];
		inv[5]=mat[2]*mat[3]-mat[0]*mat[5];
		inv[6]=mat[3]*mat[7]-mat[4]*mat[6];
		inv[7]=mat[1]*mat[6]-mat[0]*mat[7];
		inv[8]=mat[0]*mat[4]-mat[1]*mat[3];
		if (D==0) return n;

		int i;
		for (i=0;i<9;i++)
			inv[i]/=D;
//		Matrix ret=new Matrix();
		int x,y;
		for (y=0;y<3;y++)
			for (x=0;x<3;x++)
				n.m[M_CELL(x,y)]=inv[y*3+x];
//				ret.data[y][x]=inv[y*3+x];
//				ret.data[y][x]=inv[y*3+x];
		return n;
*/
	}
	inline Matrix makeScaling(Vector scale) const
	{
		Matrix	n;
//		n.m[M_CELL(0,0)] = 1.0/scale.v[0];
//		n.m[M_CELL(1,1)] = 1.0/scale.v[1];
//		n.m[M_CELL(2,2)] = 1.0/scale.v[2];
		n.m[M_CELL(0,0)] = scale.v[0];
		n.m[M_CELL(1,1)] = scale.v[1];
		n.m[M_CELL(2,2)] = scale.v[2];
		return n;
	}
protected:
	scalar	m[4*4];
};
#undef M_CELL
#endif

