#include <math.h>
#include <stdlib.h>
#include <SDL/SDL.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#define PI 3.14159265

static GLUquadric *quadric;
static GLfloat mat_specular[] = { 1.0, 1.0, 1.0, 1.0 };
static GLfloat mat_shininess[] = { 90 };
static GLfloat light_position[] = { 1, 3, 2, 0};
static GLfloat light_color[] = { 1.0, 1.0, 1.0 };
static GLfloat light_direction[] = {0.0,0.0,-1.0};
static float time = 0;
static int wave = 0;

int keyPressed()
{
	SDL_Event event;
	SDL_PollEvent(&event);
	return event.type==SDL_KEYDOWN;
}

void distord(float *x, float *y, float *z)
{
	float t = 32*time + (*y)*8;
	if (!wave) return;
	(*x) += sin(t)*.2;
	(*z) += cos(t)*.2;
}

void calcSphereVertex(float x, float y)
{
	float px = sin(x)*cos(y), py = sin(y), pz = cos(x)*cos(y);
	distord(&px,&py,&pz);
	glVertex3f(px,py,pz);
}

void renderObject()
{
	float x,y,step=(2*PI)/64;

	glEnable(GL_NORMALIZE);
	glPushAttrib(GL_CURRENT_BIT);
	glPushMatrix();
	glScalef(.5,.5,.5);
	glTranslatef(0,1,0);
	glColor3f(.3,.1,.1);
//	glutSolidTeapot(0.5);

	for(y=0;y<2*PI-step;y+=step)
	{
		for(x=0;x<2*PI-step;x+=step)
		{
			glBegin(GL_QUADS);
			calcSphereVertex(x,y);
			calcSphereVertex(x+step,y);
			calcSphereVertex(x+step,y+step);
			calcSphereVertex(x,y+step);
			glNormal3f(sin(x)*cos(y),sin(y),cos(x)*cos(y));
			glEnd();
		}
	}

	glPopMatrix();
	glPopAttrib();
	glDisable(GL_NORMALIZE);
}

void renderPlane()
{
	glPushAttrib(GL_CURRENT_BIT);
	glColor4f(.3,.3,.6,0.4);
	glBegin(GL_POLYGON);
	glVertex3f(-1.0,0,-1.0);
	glVertex3f( 1.0,0,-1.0);
	glVertex3f( 1.0,0, 1.0);
	glVertex3f(-1.0,0, 1.0);
	glEnd();
	glPopAttrib();
}

void draw()
{
	GLfloat glass_specular[] = { 0.3, 0.3, 0.6, 0.3 };
	GLfloat glass_shininess[] = { 10.0 };
	
	time+=0.01;
	
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT );
    
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity( );

	glTranslatef( 0.0, 0.0, -3.0  + sin(time));
//	glRotatef(time, 0, 0, 1.0);
	gluLookAt(0,0,0,sin(time)*4,sin(time/4.0)-2,cos(time)*4,0,1,0);

	glEnable(GL_STENCIL_TEST);
	glStencilFunc(GL_ALWAYS,1,1);
	glStencilOp(GL_REPLACE,GL_REPLACE,GL_REPLACE);

	glDepthMask(0);
	glColorMask(0,0,0,0);
	renderPlane();
	glColorMask(1,1,1,1);
	glDepthMask(1);

	glPushMatrix();
	glStencilFunc(GL_EQUAL,1,1);
	glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
//	glRotatef(180,0,0,1);
//	glRotatef(180,0,1,0);
	glScalef(1,-1,1);
	glFrontFace(GL_CCW);
	wave=1;
	renderObject();
	wave=0;
	glFrontFace(GL_CW);
	glDisable(GL_STENCIL_TEST);
	glPopMatrix();

	renderPlane();
	renderObject();
}

void init(unsigned width, unsigned height)
{
	float ratio = (float)width / (float)height;
	glShadeModel(GL_SMOOTH);
	glCullFace( GL_BACK );
	glFrontFace( GL_CW );
	glEnable( GL_CULL_FACE );
	glClearColor( 0, 0, 0, 0 );
	glViewport( 0, 0, width, height );
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity( );
	gluPerspective( 60.0, ratio, 1.0, 1024.0 );
	quadric=gluNewQuadric();

        glLightfv(GL_LIGHT0, GL_POSITION, light_position);
        glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
        glMaterialfv(GL_FRONT, GL_SHININESS, mat_shininess);
        glLightfv(GL_LIGHT0, GL_AMBIENT, light_color);

	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
//	glColorMaterial(GL_FRONT_AND_BACK,GL_AMBIENT_AND_DIFFUSE);

	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHTING);
	glEnable(GL_BLEND);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_COLOR_MATERIAL);

        glShadeModel(GL_SMOOTH);
}

int main(int argc, char **argv)
{
	SDL_Surface *screen;
//	Uint16 *b = (Uint16*)screen;
	int i;

	if (SDL_Init(SDL_INIT_AUDIO|SDL_INIT_VIDEO) < 0)
	{
	    	fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
    		exit(1);
	}
	atexit(SDL_Quit);	

/*	
	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 16 );
	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
*/	
	SDL_GL_SetAttribute( SDL_GL_STENCIL_SIZE, 8 );
//	screen = SDL_SetVideoMode(640,480,16,SDL_OPENGLBLIT);
	screen = SDL_SetVideoMode(640,480,32,SDL_OPENGL);
	
	SDL_GL_GetAttribute(SDL_GL_STENCIL_SIZE,&i);
	printf("stencil size %d.\n", i); 
	
	if (screen == NULL)
	{
		fprintf(stderr, "Unable to set 640x480 video: %s\n", SDL_GetError());
		exit(1);
	}
	
	init(640,480);

	while(!keyPressed())
	{
			draw();
			SDL_GL_SwapBuffers();
	}

	return 0;
}

